<?php
/**
 * @file
 * Module functions for moneybird commerce module.
 * Defines the functions that will be used throughout the module
 * Written by Kevin Muller (kevin@enova-tech.net),
 * Cedric Alfonsi (cedric.alfonsi@enova-tech.net)
 * Updated by Robert Ragas (robert@goalgorilla.com)
 */

/**
 * Return all the invoices contained in moneybird.
 *
 * @return array
 *   Is of type Moneybird\Invoice
 */
function _moneybird_commerce_get_all_invoices() {
  $connector = _moneybird_commerce_init_moneybird_connection();
  $moneybird = new \Picqer\Financials\Moneybird\Moneybird($connector);
  $invoices = $moneybird->salesInvoice()->get();
  return $invoices;
}

/**
 * Callback function for authorization.
 */
function moneybird_commerce_callback() {
  // If authorization code is returned from Moneybird, save this to use for token request
  if (isset($_GET['code'])) {
    variable_set('moneybird_commerce_authorizationcode', $_GET['code']);

    $redirectUrl = variable_get('moneybird_commerce_redirect_uri', '');
    $clientId = variable_get('moneybird_commerce_clientid', '');
    $clientSecret = variable_get('moneybird_commerce_clientsecret', '');

    $connection = connect($redirectUrl, $clientId, $clientSecret);

    if (is_object($connection)) {
      drupal_set_message(t('Connection success'), 'status');
      drupal_goto("admin/config/moneybird_commerce/blocks");
    } else {
      drupal_set_message($connection, 'error');
      drupal_goto("admin/config/moneybird_commerce/blocks");
    }
  }
}

/**
 * Function to connect to Moneybird, this creates the client and automatically retrieves oAuth tokens if needed
 *
 * @return \Picqer\Financials\Moneybird\Connection
 * @throws Exception
 */
function _moneybird_commerce_init_moneybird_connection() {
  // Set these params
  $redirectUrl = variable_get('moneybird_commerce_redirect_uri', '');
  $clientId = variable_get('moneybird_commerce_clientid', '');
  $clientSecret = variable_get('moneybird_commerce_clientsecret', '');
  $administrationId = variable_get('moneybird_commerce_administrationid', '');

  // Create the Moneybird client
  $connection = connect($redirectUrl, $clientId, $clientSecret);

  // Set the administration ID if access tokenis set
  if (variable_get('moneybird_commerce_accesstoken', '') != null) {
    $connection->setAdministrationId($administrationId);
  }

  return ($connection);
}

/**
 * Function to authorize with Moneybird, this redirects to Moneybird login promt and retrieves authorization code
 * to set up requests for oAuth tokens
 */
function authorize($redirectUrl, $clientId, $clientSecret)
{
  $connection = new \Picqer\Financials\Moneybird\Connection();
  $connection->setRedirectUrl($redirectUrl);
  $connection->setClientId($clientId);
  $connection->setClientSecret($clientSecret);
  $connection->redirectForAuthorization();
}

/**
 * Function to connect to Moneybird, this creates the client and automatically retrieves oAuth tokens if needed
 *
 * @return \Picqer\Financials\Moneybird\Connection
 * @throws Exception
 */
function connect($redirectUrl, $clientId, $clientSecret)
{
  $connection = new \Picqer\Financials\Moneybird\Connection();
  $connection->setRedirectUrl($redirectUrl);
  $connection->setClientId($clientId);
  $connection->setClientSecret($clientSecret);

  // Retrieves authorizationcode from database
  if (variable_get('moneybird_commerce_authorizationcode', '')) {
    $connection->setAuthorizationCode(variable_get('moneybird_commerce_authorizationcode', ''));
  }

  // Retrieves accesstoken from database
  if (variable_get('moneybird_commerce_accesstoken', '')) {
      $connection->setAccessToken(variable_get('moneybird_commerce_accesstoken', ''));
  }

  // Make the client connect and exchange tokens
  try {
    $connection->connect();
  } catch (\Exception $e) {
    return $e->getMessage();
  }

  // Save the new tokens for next connections
  variable_set('moneybird_commerce_accesstoken', $connection->getAccessToken());

  return $connection;
}

/**
 * Return the last commerce customer profile from a UID.
 *
 * @param int $uid
 *   The users ID.
 *
 * @return object
 *   Is of type commerce_customer_profile
 */
function _moneybird_commerce_get_last_commerce_customer_profile_from_uid($uid) {
  $profile_id = db_select('commerce_customer_profile', 'ccp')
      ->fields('ccp', array('profile_id'))
      ->condition('ccp.uid', $uid,'=')
      ->condition('ccp.type', 'billing', '=')
      ->orderBy('changed', 'DESC')
      ->execute()
      ->fetchField();
  $profile = commerce_customer_profile_load($profile_id);
  return $profile;
}

/**
 * Create a MoneyBird Invoice from a Commerce order.
 *
 * This function doesn't save the invoice. Only returns the object.
 *
 * @param int $order_id
 *   The order ID.
 *
 * @return bool|\Moneybird\Invoice
 *   FALSE if it didn't work or the Moneybird Invoice object.
 */
function _moneybird_commerce_get_mb_invoice_from_order($order_id) {
  $order = commerce_order_load($order_id);
  $profile = _moneybird_commerce_get_last_commerce_customer_profile_from_uid($order->uid);
  $profile = $profile->commerce_customer_address['und'][0];
  $connector = _moneybird_commerce_init_moneybird_connection();
  $moneybird = new \Picqer\Financials\Moneybird\Moneybird($connector);

  if (isset($_SESSION['tax_rate'])) {
    $taxRateId = $_SESSION['tax_rate'];
  }
  else {
    $taxRateId = variable_get('moneybird_commerce_tax_rate', '');
  }

  $salesInvoiceDetailsArray = [];
  foreach ($order->commerce_line_items['und'] as $k => $line_item) {

    /** @var \Picqer\Financials\Moneybird\Entities\SalesInvoiceDetail $salesInvoiceDetail */
    $salesInvoiceDetail = $moneybird->SalesInvoiceDetail();

    $li = commerce_line_item_load($line_item['line_item_id']);

    $product_label = '';
    if (isset($li->commerce_product[LANGUAGE_NONE][0]['product_id']) && !empty($li->commerce_product[LANGUAGE_NONE][0]['product_id'])) {
      $product = commerce_product_load($li->commerce_product[LANGUAGE_NONE][0]['product_id']);
      $product_label = $product->title;
    }

    if (empty($product_label)) {
      $product_label = $li->line_item_label;
    }

    $taxDivision = 100;

    $invoice_data = array(
                          'amount' => $li->quantity,
                          'description' => $product_label,
                          'price' => $li->commerce_unit_price['und'][0]['amount'] / $taxDivision,
                          );
    if ($taxRateId) {
      $invoice_data['taxRateId'] = $taxRateId;
    }

    $salesInvoiceDetail->price = $invoice_data['price'];
    $salesInvoiceDetail->description = $invoice_data['description'];
    $salesInvoiceDetail->amount = $invoice_data['amount'];
    $salesInvoiceDetail->tax_rate_id = $taxRateId;

    $salesInvoiceDetailsArray[] = $salesInvoiceDetail;
  }

  $mid = db_query(
    'SELECT mid FROM {moneybird_users} mu
     WHERE mu.uid =' . $order->uid
  )->fetchField();

  // If we cannot retrieve the MoneyBird Id from db (anonymous user), we
  // retrieve it from the session (it was placed there after saving).
  if (!$mid) {
    if (isset($_SESSION['moneybird_data']['contact_id'])) {
      $mid = $_SESSION['moneybird_data']['contact_id'];
    }
    else {
      // throw new Exception(t('Impossible to retrieve contact id'));
      $mid = null;

      // Create the customer.
      $name = _mb_split_full_name($profile['name_line']);
      $connector = _moneybird_commerce_init_moneybird_connection();
      $moneybird = new \Picqer\Financials\Moneybird\Moneybird($connector);
      $contact = $moneybird->contact();

      $contact->company_name = $profile['organisation_name'];
      $contact->firstname = $name['first_name'];
      $contact->lastname = $name['last_name'];
      $contact->address1 = $profile['thoroughfare'];
      $contact->address2 = $profile['premise'];
      $contact->zipcode = $profile['postal_code'];
      $contact->city = $profile['locality'];
      $contact->country = $profile['country'];
      $contact->email = $order->mail;

      try {
        $contact->save();
      } catch (Exception $e) {
        // If we cannot save the contact for some reasons, we just return. We will
        // try later.
        return FALSE;
      }

      $mid = $contact->id;
    }
  }

  /** @var \Picqer\Financials\Moneybird\Entities\Contact $contact */
  $contact = $moneybird->contact(array('id' => $mid));

  // Set workflow Id.
  if (isset($_SESSION['workflow'])) {
    $workflowId = $_SESSION['workflow'];
  }
  else {
    $workflowId = variable_get('moneybird_commerce_workflows', '');
  }

  // Set templateId.
  if (isset($_SESSION['document_style'])) {
    $templateId = $_SESSION['document_style'];
  }
  else {
    $templateId = variable_get('moneybird_commerce_documentstyle', '');
  }

  /** @var \Picqer\Financials\Moneybird\Entities\SalesInvoice $salesInvoice */
  $salesInvoice = $moneybird->salesInvoice();
  $salesInvoice->details = $salesInvoiceDetailsArray;
  $salesInvoice->prices_are_incl_tax = variable_get('moneybird_commerce_price_incl_taxes', 0);
  $salesInvoice->document_style_id = $templateId;
  $salesInvoice->workflow_id = $workflowId;
  $salesInvoice->contact_id = $contact->id;

  return $salesInvoice;
}

/**
 * Synchronize the contacts from Drupal to Moneybird.
 *
 * Basically search for the users that have no matching
 * in moneybird, then create the corresponding contact
 *
 * @param int $nb_sync
 *   Returns the number of contacts that have been synced
 *
 * @return bool
 *   TRUE if there were no contacts saved or the saving was successfull. FALSE
 *   otherwise.
 */
function _moneybird_commerce_sync_drupal_to_moneybird_contacts(&$nb_sync = 0) {
  $nb_sync = 0;

  $result = db_query(
    "SELECT uid FROM {moneybird_users} mu
     WHERE mu.mid IS NULL
     or mu.mid = ''"
  )->fetchAll();

  if (!$result) {
    return TRUE;
  }

  foreach ($result as $r) {
    $profile = _moneybird_commerce_get_last_commerce_customer_profile_from_uid($r->uid);
//    $mobilephone = $profile->field_telefonnummer_handy['und'][0]['safe_value'];
    $profile = $profile->commerce_customer_address['und'][0];
    $user = user_load($r->uid);

    // Now, either update the customer profile, or save a new one if it doesn't
    // exist.
    $name = _mb_split_full_name($profile['name_line']);
    $connector = _moneybird_commerce_init_moneybird_connection();
    $moneybird = new \Picqer\Financials\Moneybird\Moneybird($connector);
    $contact = $moneybird->contact();

    $contact->company_name = $profile['organisation_name'];
    $contact->firstname = $name['first_name'];
    $contact->lastname = $name['last_name'];
    $contact->address1 = $profile['thoroughfare'];
    $contact->address2 = $profile['premise'];
    $contact->zipcode = $profile['postal_code'];
    $contact->city = $profile['locality'];
    $contact->country = $profile['country'];
    $contact->email = $user->mail;

    try {
      $contact->save();
    } catch (Exception $e) {
      // If we cannot save the contact for some reasons, we just return. We will
      // try later.
      return FALSE;
    }

    // Update the contact mid in the table. It has now been synchronized.
    db_update('moneybird_users')
      ->fields(array(
        'mid' => $contact->id,
      ))
      ->condition('uid', $r->uid, '=')
      ->execute();

    $nb_sync++;
  }
  return TRUE;
}

/**
 * Synchronize the orders from Drupal to Moneybird.
 *
 * Basically search for the orders that have no matching in Moneybird, then
 * create the corresponding invoice
 *
 * @param int $nb_sync
 *   Returns the number of invoices that have been synced.
 *
 * @return bool
 *   TRUE if there were no orders saved or the saving was successfull. FALSE
 *   otherwise.
 */
function _moneybird_commerce_sync_drupal_to_moneybird_invoices(&$nb_sync = 0) {
  $nb_sync = 0;

  $result = db_query(
    "SELECT id FROM {moneybird_orders} mo
     WHERE mo.mid IS NULL
     or mo.mid = ''"
  )->fetchAll();

  if (!$result) {
    return TRUE;
  }

  foreach ($result as $r) {
    $order = commerce_order_load($r->id);
    $profile = $profile->commerce_customer_address['und'][0];
    $user = user_load($order->uid);

    // Now, either update the customer profile, or save a new one if it doesn't
    // exist.
    $invoice = _moneybird_commerce_get_mb_invoice_from_order($order->order_id);
    $connector = _moneybird_commerce_init_moneybird_connection();

    /** @var \Picqer\Financials\Moneybird\Entities\SalesInvoice $salesInvoice */
    $salesInvoice = $invoice;

    try {
        $salesInvoice->save();
    } catch (Exception $e) {
        // If we cannot save the invoice for some reasons, we just return. We will
        // try later.
        return FALSE;
    }

    // Update the contact mid in the table. It has now been synchronized.
      db_update('moneybird_orders')
          ->fields(array(
              'mid' => $salesInvoice->id,
          ))
          ->condition('id', $r->id, '=')
          ->execute();

    $nb_sync++;
  }
  return TRUE;
}

/**
 *
 * This function runs from a cron to check the payment status on invoices
 * changes to our application.
 *
 */
function _moneybird_commerce_sync_invoices_status(&$nb_sync = 0) {
  $nb_sync = 0;
  // First check if there are any moneybird orders who need to be checked.
  $result = db_query(
    "SELECT * FROM {moneybird_orders} mo
     WHERE mo.mid IS NOT NULL"
  )->fetchAll();

  if (!$result) {
    return TRUE;
  }

  $connector = _moneybird_commerce_init_moneybird_connection();
  $moneybird = new \Picqer\Financials\Moneybird\Moneybird($connector);

  foreach ($result as $key => $value ) {
    // Try to search the invoice otherwise go to the next foreach.
    try {
      $salesInvoice = $moneybird->salesInvoice()->find($value->mid);
    } catch (\Exception $e) {
      continue;
    }

    // If the invoice is paid change this status in Drupal commerce.
    if ($salesInvoice->state == 'paid') {
      $order = commerce_order_load($value->id);
      $order->status = 'completed';
      commerce_order_save($order);
      watchdog('moneybird', 'Updated order ' . $value->id . ' to state ' . '"completed"' . ' by push callback');
      $nb_sync++;
    }
  }
  return TRUE;
}

/**
 *
 * Function to get all the document styles defined in moneybird.
 *
 */
function _moneybird_commerce_sync_documentstyles_from_moneybird(&$nb_sync = 0) {
  $nb_sync = 0;

  $connector = _moneybird_commerce_init_moneybird_connection();
  $moneybird = new \Picqer\Financials\Moneybird\Moneybird($connector);
  $documentstyles = $moneybird->documentStyle()->get();

  foreach ($documentstyles as $key => $value ) {
    $did = db_query(
      'SELECT did FROM {moneybird_document_styles} ds
      WHERE ds.did = ' . $value->id
    )->fetchField();
    if ($did) {
      db_update('moneybird_document_styles')
        ->fields(array(
          'name' => $value->name,
        ))
        ->condition('did', $value->id, '=')
        ->execute();
    } else {
      db_insert('moneybird_document_styles')
      ->fields(array(
        'did' => $value->id,
        'name' => $value->name,
      ))
        ->execute();
      $nb_sync++;
    }
  }
  return TRUE;
}

/**
 *
 * Function to get all the taxrates defined in moneybird.
 *
 */
function _moneybird_commerce_sync_taxrates_from_moneybird(&$nb_sync = 0) {
  $nb_sync = 0;

  $connector = _moneybird_commerce_init_moneybird_connection();
  $moneybird = new \Picqer\Financials\Moneybird\Moneybird($connector);
  $taxrates = $moneybird->taxRate()->get();

  foreach ($taxrates as $key => $value ) {
    $tid = db_query(
      'SELECT tid FROM {moneybird_tax_rates} tr
      WHERE tr.tid = ' . $value->id
    )->fetchField();
    if ($tid) {
      db_update('moneybird_tax_rates')
        ->fields(array(
          'name' => $value->name,
          'tax_rate_type' => $value->tax_rate_type,
        ))
        ->condition('tid', $value->id, '=')
        ->execute();
    } else {
      db_insert('moneybird_tax_rates')
      ->fields(array(
        'tid' => $value->id,
        'name' => $value->name,
        'tax_rate_type' => $value->tax_rate_type,
      ))
        ->execute();
      $nb_sync++;
    }
  }
  return TRUE;
}

/**
 *
 * Function to get all the workflows defined in moneybird.
 *
 */
function _moneybird_commerce_sync_workflows_from_moneybird(&$nb_sync = 0) {
  $nb_sync = 0;

  $connector = _moneybird_commerce_init_moneybird_connection();
  $moneybird = new \Picqer\Financials\Moneybird\Moneybird($connector);
  $workflows = $moneybird->workflow()->get();

  foreach ($workflows as $key => $value ) {
    $wid = db_query(
      'SELECT wid FROM {moneybird_workflows} wf
      WHERE wf.wid = ' . $value->id
    )->fetchField();
    if ($wid) {
      db_update('moneybird_workflows')
        ->fields(array(
          'name' => $value->name,
          'type' => $value->type,
        ))
        ->condition('wid', $value->id, '=')
        ->execute();
    } else {
      db_insert('moneybird_workflows')
      ->fields(array(
        'wid' => $value->id,
        'name' => $value->name,
        'type' => $value->type,
      ))
        ->execute();
      $nb_sync++;
    }
  }
  return TRUE;
}


/**
 * Split a full name into a first name and a last name.
 *
 * @param string $full_name
 *   The full name that needs to be split.
 *
 * @return array
 *   With indexes "first_name" and "last_name".
 */
function _mb_split_full_name($full_name) {
  $results = array();

  $r = explode(' ', $full_name);
  $size = count($r);

  // Check first for period, assume salutation if so.
  if (mb_strpos($r[0], '.') === FALSE) {
    $results['salutation'] = '';
    $results['first_name'] = $r[0];
  }
  else {
    $results['salutation'] = $r[0];
    $results['first_name'] = $r[1];
  }

  // Check last for period, assume suffix if so.
  if (mb_strpos($r[$size - 1], '.') === FALSE) {
    $results['suffix'] = '';
  }
  else {
    $results['suffix'] = $r[$size - 1];
  }

  // Combine remains into last.
  $start = ($results['salutation']) ? 2 : 1;
  $end = ($results['suffix']) ? $size - 2 : $size - 1;

  $last = '';
  for ($i = $start; $i <= $end; $i++) {
    $last .= ' ' . $r[$i];
  }
  $results['last_name'] = trim($last);

  return $results;
}
